
/*
 * Exports for autoconfiguration
 */

#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/stat.h>
#include <sys/kmem.h>

#include "defs.h"
#include "rbuffer.h"
#include "drv.h"


/* Attaches a device and initializes device state.  */   
static int attach_device(dev_info_t* dip)
{
	int ret;
	int inst;
	devstate_t* state;


	/* Allocate soft state for the device */
	
	inst = ddi_get_instance(dip);
	ret = ddi_soft_state_zalloc(drv_soft_state, inst);
	if (ret != DDI_SUCCESS) {
		cmn_err(CE_WARN, "^opto: Soft state allocation failed!");
		return DRV_FAILURE;
	}

	state = ddi_get_soft_state(drv_soft_state, inst);
	if (state == NULL) {
		cmn_err(CE_WARN, "^opto: Soft state retrieval failed!");
		ddi_soft_state_free(drv_soft_state, inst);
		return DRV_FAILURE;
	}

	state->dip = dip;
	state->inst = inst;
	state->prev_val = 0;
	state->cur_val = ~0;
	state->intr_mask = DEFINTRMASK;

	
	/* Do not allow high-level interrupts */

	if (ddi_intr_hilevel(dip, 0) != 0) {
		cmn_err(CE_WARN, "^opto: Cannot use high-level interrupts!");
		ddi_soft_state_free(drv_soft_state, inst);
		return DRV_FAILURE;
	}


	/*  Set up the ring buffer for this device  */
	
	state->rbuf = rbuffer_create();
	if (state->rbuf == NULL) {
		cmn_err(CE_WARN, "^opto: Ring Buffer creation failed!");
		ddi_soft_state_free(drv_soft_state, inst);
		return DRV_FAILURE;		
	}

	ret = state->rbuf->construct(state->rbuf, RBUF_DEF_SIZE);
	if (ret != 0) {
		cmn_err(CE_WARN, "^opto: Ring buffer construction failed!");
		rbuffer_destroy(state->rbuf);
		ddi_soft_state_free(drv_soft_state, inst);
		return DRV_FAILURE;
	}


	/*  Map device registers  */
	
	ret = ddi_regs_map_setup(dip, 0, (caddr_t*)&(state->regs), 0, 0, &drv_acc_attr, &(state->ah));
	if (ret != DDI_SUCCESS) {
		cmn_err(CE_WARN, "^opto: Register mapping failed!");
		rbuffer_destroy(state->rbuf);
		ddi_soft_state_free(drv_soft_state, inst);
		return DRV_FAILURE;
	
	}	

	pic8259a_init(state, OPTO_LTIM_FLANK);
	cnt8254_init(state, 0, CNT0_DEFMODE, CNT0_DEFSVAL);
	cnt8254_init(state, 1, CNT1_DEFMODE, CNT1_DEFSVAL);
	cnt8254_init(state, 2, CNT2_DEFMODE, CNT2_DEFSVAL);


	/*  Set up the interrupt handling  */
	
	ret = ddi_get_iblock_cookie(dip, 0, &(state->ibc));
	if (ret != DDI_SUCCESS) {
		cmn_err(CE_WARN, "^opto: Cannot get iblock cookie!");
		ddi_regs_map_free(&state->ah);
		rbuffer_destroy(state->rbuf);
		ddi_soft_state_free(drv_soft_state, inst);
		return DRV_FAILURE;
	}
	
	mutex_init(&(state->mtx), NULL, MUTEX_DRIVER, state->ibc);

	ret = ddi_add_intr(dip, 0, &(state->ibc), NULL, drv_intrh, (caddr_t)inst);
	if (ret != DDI_SUCCESS) {
		cmn_err(CE_WARN, "^opto: Cannot set up interrupt handler!");
		mutex_destroy(&(state->mtx));
		ddi_regs_map_free(&state->ah);
		rbuffer_destroy(state->rbuf);
		ddi_soft_state_free(drv_soft_state, inst);
		return DRV_FAILURE;
	}


	/*  Create the minor node  */

	ret = ddi_create_minor_node(dip, ddi_get_name(dip), S_IFCHR, inst, DRV_NODE_TYPE, 0);
	if (ret != DDI_SUCCESS) {
		cmn_err(CE_WARN, "^opto: Minor node creation failed!");
		ddi_remove_intr(dip, 0, state->ibc);
		mutex_destroy(&(state->mtx));
		ddi_regs_map_free(&state->ah);
		rbuffer_destroy(state->rbuf);
		ddi_soft_state_free(drv_soft_state, inst);
		return DRV_FAILURE;
	}


	/*  Report the device and finish  */
	
	ddi_report_dev(dip);
	return DRV_SUCCESS;
}


/*  Detaches the device and frees resources allocated
    for it. Basically, it does the reverse of what attach_device() did.   */
static int detach_device(dev_info_t* dip)
{
	int inst;
	devstate_t* state = NULL;

	inst = ddi_get_instance(dip);
	state = ddi_get_soft_state(drv_soft_state, inst);
	
	ddi_remove_minor_node(dip, ddi_get_name(dip));

	if (state != NULL) {
		cnt8254_init(state, 2, OPTO_CNTMODE_SOFTWARE, 0);
		pic8259a_set_intr_mask(state, 0x0);

		ddi_remove_intr(dip, 0, state->ibc);
		mutex_destroy(&state->mtx);
		ddi_regs_map_free(&state->ah);
		rbuffer_destroy(state->rbuf);
	}
	
	ddi_soft_state_free(drv_soft_state, inst);
	return DRV_SUCCESS;
}


/*  The attach DDI routine.
    see 'man attach' for more info.  */
int drv_attach(dev_info_t* dip, ddi_attach_cmd_t cmd)
{
	int ret = DDI_FAILURE;

	switch (cmd) {
	case DDI_ATTACH:
		if (attach_device(dip) == DRV_SUCCESS)
			ret = DDI_SUCCESS;
		break;
	}

	return ret;
}


/*  The detach routine. 
    see 'man detach' for more info.  */
int drv_detach(dev_info_t* dip, ddi_detach_cmd_t cmd)
{
	int ret = DDI_FAILURE;

	switch (cmd) {
	case DDI_DETACH:
		if (detach_device(dip) == DRV_SUCCESS)
			ret = DDI_SUCCESS;
	}

	return ret;
}


/*  The probe routine.
    see 'man probe' for more info.  */
int drv_probe(dev_info_t* dip)
{
	return DDI_PROBE_SUCCESS;
}


/*  The getinfo routine.
    see 'man getinfo' for more info.  */
int drv_getinfo(dev_info_t* dip, ddi_info_cmd_t cmd, void* arg, void** res)
{
	int ret = DDI_FAILURE;
	devstate_t* state;

	*res = NULL;

	switch (cmd) {
	case DDI_INFO_DEVT2INSTANCE:
		*res = (void*)getminor((dev_t)arg);
		ret = DDI_SUCCESS;
		break;

	case DDI_INFO_DEVT2DEVINFO:
		state = ddi_get_soft_state(drv_soft_state, getminor((dev_t)arg));
		if (state != NULL) {
			*res = state->dip;
			ret = DDI_SUCCESS;
		}
	}

	return ret;
}


