
/*
 * Charracter device IO operations
 */

#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/poll.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/kmem.h>

#include "defs.h"
#include "rbuffer.h"
#include "drv.h"


/*  Opens the device.  */
int drv_open(dev_t* dev, int oflags, int otype, cred_t* cred)
{
	int inst;
	devstate_t* state;

	if (otype != OTYP_CHR) {
		return (EINVAL);
	}

	inst = getminor(*dev);
	state = ddi_get_soft_state(drv_soft_state, inst);
	if (state != NULL) {
		mutex_enter(&state->mtx);

		if (state->refc == 0) {
			pic8259a_set_intr_mask(state, state->intr_mask);
		}

		state->refc++;

		mutex_exit(&state->mtx);
	}

	return (0);
}


/*  Closes the device  */
int drv_close(dev_t dev, int oflags, int otype, cred_t* cred)
{
	int inst;
	devstate_t* state;

	inst = getminor(dev);
	state = ddi_get_soft_state(drv_soft_state, inst);
	if (state != NULL) {
		mutex_enter(&state->mtx);
		state->refc--;
		
		if (state->refc == 0) {
			pic8259a_set_intr_mask(state, 0x0);
		}
		
		mutex_exit(&state->mtx);
	}

	return (0);
}


/*  Reads some bytes from the device. Device must have
    been opened for this to work :-)  */
int drv_read(dev_t dev, struct uio* uiop, cred_t* cred)
{
	int ret, inst;
	byte_t* buf;
	int buf_size;
	int bts_read;
	devstate_t* state;

	inst = getminor(dev);
	state = ddi_get_soft_state(drv_soft_state, inst);
	if (state == NULL) {
		return (ENXIO);
	}
	
	if (state->rbuf->stored == 0) {
		return (0);
	}
	
	mutex_enter(&state->mtx);
	
	buf_size = min(uiop->uio_resid, state->rbuf->stored);
	
	buf = (byte_t*)kmem_zalloc(buf_size, KM_NOSLEEP);
	if (buf != NULL) {
		bts_read = state->rbuf->read(state->rbuf, buf, buf_size);
		ret = uiomove(buf, bts_read, UIO_READ, uiop);
		kmem_free(buf, buf_size);
	} else {
		ret = ENOMEM;
	}

	mutex_exit(&state->mtx);
		
	return (ret);
}


/*  Writes, or better attempts to write to the device.  */
int drv_write(dev_t dev, struct uio* uiop, cred_t* cred)
{
	int ret, inst;
	byte_t* buf, *bufwalk;
	int buf_size;
	int bts_wrote;
	devstate_t* state;

	inst = getminor(dev);
	state = ddi_get_soft_state(drv_soft_state, inst);
	if (state == NULL) {
		return (ENXIO);
	}
	
	mutex_enter(&state->mtx);
	
	buf_size = min(uiop->uio_resid, state->rbuf->capacity);
	
	buf = (byte_t*)kmem_zalloc(buf_size, KM_NOSLEEP);
	if (buf != NULL) {
		ret = uiomove(buf, buf_size, UIO_WRITE, uiop);
		if (ret == 0) {
			bufwalk = buf;
			while ((bufwalk - buf) < buf_size)
				state->rbuf->store(state->rbuf, *bufwalk++);
		}
		kmem_free(buf, buf_size);
	} else {
		ret = ENOMEM;
	}

	mutex_exit(&state->mtx);
		
	if (ret == 0)
		pollwakeup(&state->ph, POLLIN);

	return (ret);
}


/*  Performs multiplexing IO on the device.  */
int drv_poll(dev_t dev, short events, int anyyet, 
	     short* revents, struct pollhead** ph)
{
	short re = 0;
	devstate_t* state;

	state = ddi_get_soft_state(drv_soft_state, getminor(dev));
	if (state == NULL)
		return (ENXIO);

	if ((events & POLLIN) && (state->rbuf->stored > 0))
		re |= POLLIN;
	
	if (re == 0) {
		if (!anyyet)
			*ph = &state->ph;
	}

	*revents = re;
	return (0);
}

