
/*
 * Device I/O
 *
 */

#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "drv.h"


/*  ioctl entry point for userland applications who know
    what they are doing.  */
int drv_ioctl(dev_t dev, int cmd, intptr_t arg, 
	      int mode, cred_t* cred, int* rval)
{
	int ret = 0;
	byte_t b;
	opto_reg_t ori;
	opto_cnt_t oci;
	devstate_t* state;

	state = ddi_get_soft_state(drv_soft_state, getminor(dev));
	if (state == NULL)
		return (ENXIO);

	switch (cmd) {
	case OPTOCTL_GET_INTRMASK:
		ret = ddi_copyout(&state->intr_mask, (void*)arg, sizeof(byte_t), mode);
		break;
	
	case OPTOCTL_SET_INTRMASK:
		ret = ddi_copyin((void*)arg, &b, sizeof(byte_t), mode);
		if (ret == 0) {
			mutex_enter(&state->mtx);
			state->intr_mask = b;
			pic8259a_set_intr_mask(state, state->intr_mask);
			mutex_exit(&state->mtx);
		} else {
			ret = EFAULT;
		}
		break;
	
	case OPTOCTL_READ_REG:
		ret = ddi_copyin((void*)arg, &ori, sizeof(opto_reg_t), mode);
		if (ret == 0) {
			if (OPTO_REG_IS_VALID(ori.reg)) {
				mutex_enter(&state->mtx);
				ori.val = ddi_get8(state->ah, &(state->regs->in00_07) + ori.reg);
				mutex_exit(&state->mtx);
			}
			ret = ddi_copyout(&ori, (void*)arg, sizeof(opto_reg_t), mode);
		}
		if (ret != 0)
			ret = EFAULT;
		break;
	
	case OPTOCTL_WRITE_REG:
		ret = ddi_copyin((void*)arg, &ori, sizeof(opto_reg_t), mode);
		if (ret == 0) {
			if (OPTO_REG_IS_VALID(ori.reg)) {
				mutex_enter(&state->mtx);
				ddi_put8(state->ah, &(state->regs->in00_07) + ori.reg, ori.val);
				mutex_exit(&state->mtx);
			}
		} else
			ret = EFAULT;
		break;

	case OPTOCTL_CNT_INIT:
		ret = ddi_copyin((void*)arg, &oci, sizeof(opto_cnt_t), mode);
		if (ret == 0) {
			if (OPTO_CNT_IS_VALID(oci.num)) {
				mutex_enter(&state->mtx);
				pic8259a_set_intr_mask(state, 0x0);
				cnt8254_init(state, oci.num, oci.mode, (word_t)oci.stval);
				pic8259a_set_intr_mask(state, state->intr_mask);
				mutex_exit(&state->mtx);
			}
		} else
			ret = EFAULT;
		break;

	case OPTOCTL_PIC_INIT:
		ret = ddi_copyin((void*)arg, &b, sizeof(byte_t), mode);
		if (ret == 0) {
			mutex_enter(&state->mtx);
			pic8259a_set_intr_mask(state, 0x0);
			pic8259a_init(state, b);
			pic8259a_set_intr_mask(state, state->intr_mask);
			mutex_exit(&state->mtx);
		} else
			ret = EFAULT;
		break;

	default:
		ret = ENOTTY;
	}

	return (ret);
}



/*  Initializes the on-board programmable interrupt
    controller PIC 8259A of OPTOIN-16 Extended card. The LTIM
    (level trigger mode) can be specified.
    
    If the PIC is being reinitialized at runtime, interrupts
    should be disabled before doing it.  */
void pic8259a_init(devstate_t* state, byte_t ltim)
{
	byte_t ICW1;

	ICW1 = (0x1 << 1) | (ltim << 3);
	
	ddi_put8(state->ah, &(state->regs->irrisr), ICW1);
	ddi_put8(state->ah, &(state->regs->imr), 0xff);
}


/*  Sets the interrupt mask. This selects a subset of seven inputs
    which can generate an interrupt. Since there is more logic in
    setting the activated channel's bit to one, we will make that
    the parameter to this function. The hardware wants is otherwise.  */
void pic8259a_set_intr_mask(devstate_t* state, byte_t im)
{
	ddi_put8(state->ah, &(state->regs->imr), ~im);
}


/*  Initializes the on-board timer CNT 8254 found on
    the OPTOIN-16 Extended card.  */
void cnt8254_init(devstate_t* state, byte_t tnum, byte_t tmode, word_t stval)
{
	byte_t CTRL = (tnum << 6) | (0x3 << 4) | (tmode << 1);
	byte_t* cnt = &(state->regs->cnt0) + tnum;
	
	ddi_put8(state->ah, &(state->regs->ctrl), CTRL);
	ddi_put8(state->ah, cnt, LOBYTE(stval));
	ddi_put8(state->ah, cnt, HIBYTE(stval));
}


