
/* 
 * Ring Buffer implementation
 *
 */

#include <sys/conf.h>
#include <sys/kmem.h>

#include "defs.h"
#include "rbuffer.h"


/*  Constructs the existing instance of the rbuffer type.
    size parameter must contain the number of bytes to 
    allocate for the buffer and must be greater than zero.  */
static int rbuffer_construct(rbuffer_t* rb, size_t size)
{
	int ret = (-1);

	rb->begin = (byte_t*)kmem_alloc(size, KM_NOSLEEP);
	if (rb->begin != NULL) {
		rb->end = rb->begin + (size-1);
		rb->readp = rb->begin;
		rb->writep = rb->begin;
    
		rb->stored = 0;
		rb->capacity = size;

		ret = 0;
	}

	return (ret);
}


/*  Destructs the instance of rbuffer. The instance must
    have been constructed before it can be destructed.
    Note that this does not destroy the rbuffer, but merely
    frees resources used by it internally.  */
static void rbuffer_destruct(rbuffer_t* rb)
{
	kmem_free(rb->begin, rb->capacity);
  
	rb->begin = NULL;
	rb->end = NULL;
	rb->readp = NULL;
	rb->writep = NULL;

	rb->capacity = 0;
	rb->stored = 0;
}


/*  Reads the specified amount of bytes from the rbuffer
    and removes those bytes that have been read. Return
    value is the actual number of bytes read.  */
static int rbuffer_read(rbuffer_t* rb, byte_t* val, int count)
{
	int ret = 0;

	if (count > rb->stored)
		count = rb->stored;
	ret = count;

	while (count > 0) {
		*val++ = *(rb->readp++);
		if (rb->readp > rb->end)
			rb->readp = rb->begin;
		rb->stored--;
		count--;
	}

	return (ret - count);
}


/*  Stores a byte in rbuffer. When the buffer is full,
    the oldest entry is discarded.  */
static void rbuffer_store(rbuffer_t* rb, byte_t val)
{
	*(rb->writep++) = val;
	if (rb->writep > rb->end)
		rb->writep = rb->begin;

	if (rb->stored < rb->capacity)
		rb->stored++;
	else if (++(rb->readp) > rb->end)
		rb->readp = rb->begin;
}


/*  Creates an uninitialized instance of the rbuffer type.
    rbuffer::construct() must be called to initialize the
    instance before it can be used.  */
rbuffer_t* rbuffer_create(void)
{
	rbuffer_t* rb;

	rb = (rbuffer_t*) kmem_zalloc(sizeof(rbuffer_t), KM_NOSLEEP);
	if (rb != NULL) {
		rb->construct = rbuffer_construct;
		rb->destruct = rbuffer_destruct;
		rb->read = rbuffer_read;
		rb->store = rbuffer_store;
	}

	return rb;
}


/*  Destroys the rbuffer instance. rbuffer::destruct should, but
    does not have to be called before this one.  */
void rbuffer_destroy(rbuffer_t* rb)
{
	if (rb->begin != NULL)
		rb->destruct(rb);

	kmem_free(rb, sizeof(rbuffer_t));
}





























