
/*
 *  optoctl utility
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "opto.h"


#define DEFDEV "/dev/opto0"

#define DO_NOTHING 0x0
#define DO_CNT     0x1
#define DO_REG     0x2
#define DO_IM      0x4
#define DO_PIC     0x8
#define DO_INFO    0x10
#define DO_HELP    0x20

typedef struct action {
	int todo;
	char* device;
	unsigned char ltim;
	unsigned char im;
	opto_reg_t* regs;
	int numregs;
	opto_cnt_t* cnts;
	int numcnts;
} action_t;


unsigned char decode_reg(char* str)
{
	if (strcasecmp(str, "in0") == 0)
		return OPTOIN16_REG_IN0;
	else if (strcasecmp(str, "in1") == 0)
		return OPTOIN16_REG_IN1;
	else if (strcasecmp(str, "cnt0") == 0)
		return OPTOIN16_REG_CNT0;
	else if (strcasecmp(str, "cnt1") == 0)
		return OPTOIN16_REG_CNT1;
	else if (strcasecmp(str, "cnt2") == 0)
		return OPTOIN16_REG_CNT2;
	else if (strcasecmp(str, "ctrl") == 0)
		return OPTOIN16_REG_CTRL;
	else if (strcasecmp(str, "irrisr") == 0)
		return OPTOIN16_REG_IRRISR;
	else if (strcasecmp(str, "imr") == 0)
		return OPTOIN16_REG_IMR;
	else
		return (0xff);
}


/*  Display help screen  */
void display_help(void)
{
	char help[] =
		"\n"
		"optoctl utility, version 1.0\n"
		"calling syntax:\n\n"
		"optoctl [-i] [-h] [-p ltim] [-c num:mode:sval]\n"
		"        [-r reg:rval] [-m imask] [dev]\n\n"
		"-i                Display curent device settings.\n"
		"-h                Display this help screen.\n"
		"-p ltim           Set PIC to the operating mode 'ltim'.\n"
		"-c num:mode:sval  Set counter 'num' to the operating\n"
		"                  mode 'mode' with start value 'sval'.\n"
		"-r reg:rval       Set register 'reg' to 'rval'.\n"
		"-m imask          Set interrupt mask to 'imask'.\n"
		"dev               Device to operate on (defaults to\n"
		"                  '/dev/opto0').\n\n";

	printf(help);
	return;
}


/*  Display information about the device  */
void display_dev_info(int fd)
{
	int ret;
	unsigned char imask = 0;
	opto_reg_t ri = {0, '?'};

	ret = opto_get_intrmask(fd, &imask);
	printf("Intr. Mask: %02x\n", imask);
	printf("------------------------------------------\n");

	/* input registers */

	ri.reg = OPTOIN16_REG_IN0;
	ret = opto_read_reg(fd, &ri);
	printf("Inputs:     IN0.....%02x", ri.val);
	
	ri.reg = OPTOIN16_REG_IN1;
	ret = opto_read_reg(fd, &ri);
	printf("  IN1...%02x\n", ri.val);

	/* counters */

	ri.reg = OPTOIN16_REG_CNT0;
	ret = opto_read_reg(fd, &ri);
	printf("Timers:     CNT0....%02x", ri.val);
	
	ri.reg = OPTOIN16_REG_CNT1;
	ret = opto_read_reg(fd, &ri);
	printf("  CNT1..%02x", ri.val);

	ri.reg = OPTOIN16_REG_CNT2;
	ret = opto_read_reg(fd, &ri);
	printf("  CNT2..%02x\n", ri.val);

	/* control register */
	
	ri.reg = OPTOIN16_REG_CTRL;
	ret = opto_read_reg(fd, &ri);
	printf("Control:    CTRL....%02x\n", ri.val);

	/* PIC */
	
	ri.reg = OPTOIN16_REG_IRRISR;
	ret = opto_read_reg(fd, &ri);
	printf("PIC:        IRRISR..%02x", ri.val);
	
	ri.reg = OPTOIN16_REG_IMR;
	ret = opto_read_reg(fd, &ri);
	printf("  IMR...%02x\n\n", ri.val);

	return;
}


/* Creates new action structure  */
action_t* action_create(void)
{
	action_t* act;

	act = (action_t*)malloc(sizeof(action_t));
	if (act != NULL) {
		act->todo = DO_NOTHING;
		act->device = DEFDEV;
		act->ltim = OPTO_LTIM_FLANK;
		act->im = 0x0;
		act->regs = NULL;
		act->cnts = NULL;
		act->numregs = act->numcnts = 0;
	}

	return (act);
}

/* destroys the action */
void action_destroy(action_t* act)
{
	if (act->regs != NULL)
		free(act->regs);

	if (act->cnts != NULL)
		free(act->cnts);

	free(act);
	return;
}

/* adds new register action */
int action_add_reg(action_t* act, unsigned char reg, unsigned char val)
{
	if (act->regs == NULL)
		act->regs = (opto_reg_t*)malloc(sizeof(opto_reg_t));
	else
		act->regs = (opto_reg_t*)realloc(act->regs, (act->numregs+1)*sizeof(opto_reg_t));
	
	if (act->regs != NULL) {
		act->regs[act->numregs].reg = reg;
		act->regs[act->numregs].val = val;
		act->numregs++;
	}

	return (act->regs != NULL? 0 : -1);
}

/* adds new counter action */
int action_add_cnt(action_t* act, unsigned char cnt, unsigned char mode, int sval)
{
	if (act->cnts == NULL)
		act->cnts = (opto_cnt_t*)malloc(sizeof(opto_cnt_t));
	else
		act->cnts = (opto_cnt_t*)realloc(act->cnts, (act->numcnts+1)*sizeof(opto_cnt_t));
	
	if (act->cnts != NULL) {
		act->cnts[act->numcnts].num = cnt;
		act->cnts[act->numcnts].mode = mode;
		act->cnts[act->numcnts].stval = sval;
		act->numcnts++;
	}

	return (act->cnts != NULL? 0 : -1);
}


/* performs an action */
int action_perform(action_t* act)
{
	int ret = -1;
	int fd;
	int i;
	
	if (act->todo == DO_HELP || act->todo == DO_NOTHING) {
		display_help();
		return (0);
	}

	fd = open(act->device, O_RDWR);
	if (fd == (-1)) {
		printf("optoctl: Unable to open device %s\n", act->device);
		return (ret);
	}
	
	if (act->todo & DO_PIC) {
		ret = opto_pic_init(fd, &act->ltim);
		printf("optoctl: initialized PIC with LTIM=%02x. response:%x\n", act->ltim, ret);
	}

	if (act->todo & DO_IM) {
		ret = opto_set_intrmask(fd, &act->im);
		printf("optoctl: set interrupt mask to %02x. response:%x\n", act->im, ret);
	}
	
	if (act->todo & DO_CNT) {
		for (i = 0; i < act->numcnts; i++) {
			ret = opto_cnt_init(fd, &(act->cnts[i]));
			printf("optoctl: set counter %02x to mode %02x, svalue %04x. response:%x\n", 
			       act->cnts[i].num, act->cnts[i].mode, act->cnts[i].stval, ret);
		}
	}

	if (act->todo & DO_REG) {
		for (i = 0; i < act->numregs; i++) {
			ret = opto_write_reg(fd, &(act->regs[i]));
			printf("optoctl: set register %02x to %02x. response:%x\n", 
			       act->regs[i].reg, act->regs[i].val, ret);
		}
	}

	if (act->todo & DO_INFO) {
		printf("\nDevice:     %s\n", act->device);
		display_dev_info(fd);
		ret = 0;
	}

	close(fd);
	return ret;
}




/*  Main entry point  */
int main(int argv, char** argc)
{
	int i;
	int pstate = DO_NOTHING;
	char* walk, *chp;
	unsigned char cnt, mode, reg, val;
	int sval;
	action_t* act;

	act = action_create();
	if (act == NULL)
		return (127);

	/* parse command line */
	for (i = 1; i < argv; i++) {
		walk = argc[i];

		if (act->todo == DO_HELP)
			break;

		if (pstate == DO_PIC) {
			act->ltim = (*walk == '0'? OPTO_LTIM_HIGH : OPTO_LTIM_FLANK);
			act->todo |= DO_PIC;
			pstate = 0;

		} else if (pstate == DO_CNT) {
			chp = walk;
			while (*chp != '\0') {
				if (*chp == ':')
					*chp = '\0';
				chp++;
			}
			cnt = (unsigned char)strtol(walk, NULL, 0);
			while (*walk != '\0') walk++; walk++;
			mode = (unsigned char)strtol(walk, NULL, 0);
			while (*walk != '\0') walk++; walk++;
			sval = (int)strtol(walk, NULL, 0);
			
			if (cnt > 0x2 || mode > 0x5 || sval > 0xffff)
				act->todo = DO_HELP;

			action_add_cnt(act, cnt, mode, sval);
			act->todo |= DO_CNT;
			pstate = 0;

		} else if (pstate == DO_REG) {
			chp = walk;
			while (*chp != '\0') {
				if (*chp == ':')
					*chp = '\0';
				chp++;
			}
			reg = decode_reg(walk);
			while (*walk != '\0') walk++; walk++;
			val = (unsigned char)strtol(walk, NULL, 0);
			
			if (reg == 0xff || val > 0xff)
				act->todo = DO_HELP;

			action_add_reg(act, reg, val);
			act->todo |= DO_REG;
			pstate = 0;
		
		} else if (pstate == DO_IM) {
			act->im = (unsigned char)strtol(walk, NULL, 0);
			act->todo |= DO_IM;
			pstate = 0;
		
		} else {

			while (*walk == '-')
				walk++;
		
			switch (*walk) {
			case 'i':
				pstate = 0;
				act->todo |= DO_INFO;
				break;
			case 'h':
				pstate = 0;
				act->todo = DO_HELP;
				break;
			case 'p':
				pstate = DO_PIC;
				break;
			case 'c':
				pstate = DO_CNT;
				break;
			case 'r':
				pstate = DO_REG;
				break;
			case 'm':
				pstate = DO_IM;
				break;
			case '/':
				pstate = 0;
				act->device = walk;
				break;
			default:
				act->todo = DO_HELP;
				pstate = 0;
			}
		}
	}

	action_perform(act);

	action_destroy(act);
	return (0);
}
